/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Suffix Array based String Kernel.
 *
 * The Initial Developer of the Original Code is
 * Statistical Machine Learning Program (SML), National ICT Australia (NICTA).
 * Portions created by the Initial Developer are Copyright (C) 2006
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 *   Choon Hui Teo <ChoonHui.Teo@rsise.anu.edu.au>
 *   S V N Vishwanathan <SVN.Vishwanathan@nicta.com.au>
 *
 * ***** END LICENSE BLOCK ***** */


// File    : sask/Code/W_msufsort.cpp
//
// Authors : Choon Hui Teo      (ChoonHui.Teo@rsise.anu.edu.au)
//           S V N Vishwanathan (SVN.Vishwanathan@nicta.com.au)
//
// Created : 09 Feb 2006
//
// Updated : 24 Apr 2006
//           13 Jul 2007  : use MSufSort v3.1 instead of v2.2

//' Wrapper for Michael Maniscalco's MSufSort version 3.1 algorithm
#ifndef W_MSUFSORT_CPP
#define W_MSUFSORT_CPP

#include <iostream>
#include <cstring>
#include <cassert>

#include "W_msufsort.h"


W_msufsort::W_msufsort()
{
	msuffixsorter = new MSufSort();
}

W_msufsort::~W_msufsort()
{
	delete msuffixsorter;
}


/**
 *  Construct Suffix Array using Michael Maniscalco's algorithm
 *
 *  \param text     - (IN)  The text which resultant SA corresponds to.
 *  \param text_len - (IN)  The length of the text.
 *  \param sarray   - (OUT) Suffix array (memory preallocated).
 */
ErrorCode
W_msufsort::ConstructSA(SYMBOL *text, const UInt32 &text_len, UInt32 *&sarray)
{
	//printf("in ConstructSA\n");
    
    //printf("text (%d) : \"%s\"\n",text_len, text);
    // [1054:130707]
    // note that in MSufSort v3.1, the input (and output) is an array of (A+4+B)*n bytes 
    // where A := sizeof(SYMBOL)
    //       B := sizeof(UInt32)
    //       n := the number of character in the text
    
	// A temporary copy of text and SA
    // Assumption: 1 unsigned char == 1 byte
    
    int nBytes = sizeof(SYMBOL)*text_len;
    unsigned char *text_and_sarray = new unsigned char[(nBytes+4)*6];  // as required
    int *sarray_ptr = (int*)(text_and_sarray+((nBytes + 3) & ~3));     // refer to Demo.cpp in MSufSort v3.1.1
    

	assert(text_and_sarray != NULL);

    //printf("debug 1\n");

	memcpy(text_and_sarray, text, nBytes);
	msuffixsorter->Sort(text_and_sarray, nBytes);
    
    //printf("debug 2\n");
    
    for(UInt32 i=0; i<text_len; i++) {
        sarray[i] = sarray_ptr[i+1];    // skip the conceptual end-of-string character
        //printf("sa[%d] : %d  '%s'\n",i,sarray[i],&text[sarray[i]]);
        //printf("sa[%d] : %d\n",i,sarray[i]);
    }

// 	//' Code adapted from MSufSort::verifySort()	
// 	for (UInt32 i = 0; i < text_len; i++) {
//         //UInt32 tmp = msuffixsorter->ISA(i)-1;
//         //UInt32 tmp = msuffixsorter->GetValue(&i)-1;
//         sarray[tmp] = i;
//         printf("sa[%d] : %d\n",tmp,i);
// 	}

    //printf("debug 3\n");
    
	//' Deallocate the memory allocated for #text_and_sarray#
	delete [] text_and_sarray;
	
	return NOERROR;
}
#endif



