#include <fstream>
#include "DataType.h"
#include "ErrorCode.h"
#include "ESA.h"
#include "StringKernel.h"
#include <sys/time.h>
#include <cmath>

using namespace std;



void help()
{
        printf("==========================================\n"
               "=    Compute single kernel evaluation    =\n"
               "==========================================\n"
               "Usage: ./singlek  swf  param  string1  string2\n"
               "Options:\n"
               "swf (substring weighting functions)\n"
               "  0 - Constant Weight (any dummy parameter)\n"
               "  1 - Exponential Decay (lambda=0.5)\n"
               "  2 - K-Spectrum (k=5)\n"
               "  3 - Bounded Range (n=1)\n"
               "param (swf parameter)\n\n"
               "Note: \n"         
               "1. A SENTINEL character is appended to each input string.)\n\n"
              );
}


int main(int argc, char *argv[])
{
        int swf = atoi(argv[1]);
        Real swfParam = atof(argv[2]);
        string s1 = string(argv[3]);
        string s2 = string(argv[4]);
	Real k1=0.0, k2=0.0, k3=0.0;
      
        string tmps1 = s1+SENTINEL;
        string tmps2 = s2+SENTINEL;

        // setup string kernel
        StringKernel sk1(tmps1.size(), (SYMBOL*)tmps1.c_str(), swf, swfParam);
        sk1.Set_Lvs();
        sk1.PrecomputeVal();

        StringKernel sk2(tmps2.size(), (SYMBOL*)tmps2.c_str(), swf, swfParam);
        sk2.Set_Lvs();
        sk2.PrecomputeVal();

        // kernel evaluations
        sk1.Compute_K((SYMBOL*)s1.c_str(), s1.size(), k1);
        sk2.Compute_K((SYMBOL*)s2.c_str(), s2.size(), k2);
        sk2.Compute_K((SYMBOL*)s1.c_str(), s1.size(), k3); 


        // dump output
        cout << "s1       : '" << s1 << "'" << endl;
        cout << "s2       : '" << s2 << "'" << endl;
        cout << "swf      : " << swf << endl;
        cout << "swf param: " << swfParam << endl;
        cout << "k(s1,s1) : " << k1 << endl;
        cout << "k(s2,s2) : " << k2 << endl;
        cout << "k(s1,s2) : " << k3 << endl;
        cout << "normalised k(s1,s2): " << k3/sqrt(k1*k2) << endl;

	return EXIT_SUCCESS;
}
